import requests
import geocoder
from openai import OpenAI

client = OpenAI()

temperature = 0.6
basic_model = "gpt-3.5-turbo-16k"


def geocoder_api(query):
    g = geocoder.geonames(query, key='eliaswf', maxRows=1)
    return g.lat, g.lng


def summarize(query:str, content: str) -> str:
    """
    Summarizes a piece of text using the OpenAI API.

    :param query: The query to summarize.
    :type query: str
    :param content: The text to summarize.
    :type content: str
    :return: The summary.
    :rtype: str
    """
    response = client.chat.completions.create(model=basic_model,
    messages=[{'role': 'system', 'content': f'There was a search for the following weather:\n"{query}"\nPlease '
                                            f'provide a concise summary of the following content while keeping '
                                            f'mind what will best respond to the query:\n{content}\n'}],
    max_tokens=400,
    n=1,
    stop=None,
    temperature=temperature)
    summary = response.choices[0].message.content
    return summary


def get_weather(city_name):
    api_key = '916a78d6305cef8f326831938dfe03f7'
    lat, lng = geocoder_api(city_name)
    url = f"https://api.openweathermap.org/data/2.5/forecast?lat={lat}&lon={lng}&appid={api_key}&units=imperial"

    response = requests.get(url)

    if response.status_code == 200:
        weather_data = response.json()
        formatted_response = ""

        # Extract the city information
        city = weather_data['city']['name']
        country = weather_data['city']['country']
        formatted_response += f"Weather Forecast for {city}, {country}\n\n"

        # Go through each weather entry in the list
        for entry in weather_data['list']:
            # Convert temperature from Kelvin to Celsius
            temp_farenheit = entry['main']['temp']
            feels_like_farenheit = entry['main']['feels_like']
            temp_min_farenheit = entry['main']['temp_min']
            temp_max_farenheit = entry['main']['temp_max']

            # Format the date and time
            formatted_date = entry['dt_txt']

            # Add the details to the response
            formatted_response += f"{formatted_date}\n"
            formatted_response += f"   - Temperature: {temp_farenheit:.2f}°F (Feels like: {feels_like_farenheit:.0f}°F)\n"
            formatted_response += f"   - Min Temperature: {temp_min_farenheit:.0f}°F\n"
            formatted_response += f"   - Max Temperature: {temp_max_farenheit:.0f}°F\n"
            formatted_response += f"   - Pressure: {entry['main']['pressure']} hPa\n"
            formatted_response += f"   - Humidity: {entry['main']['humidity']}%\n"
            formatted_response += f"   - Weather: {entry['weather'][0]['description'].capitalize()}\n"
            formatted_response += f"   - Cloudiness: {entry['clouds']['all']}%\n"
            formatted_response += f"   - Wind: {entry['wind']['speed']} m/s, {entry['wind']['deg']} degrees\n"
            if 'rain' in entry:
                formatted_response += f"   - Rain Volume: {entry['rain']['3h']} mm/3h\n"
            formatted_response += f"   - Probability of Precipitation: {entry['pop'] * 100}%\n\n"

        #return summarize(city_name, formatted_response)
        return formatted_response
    else:
        return "City not found or request failed"



import json
import multiprocessing
import threading
import requests
from bs4 import BeautifulSoup
from openai import OpenAI, _utils
from googlesearch import search as google_search
from tiktoken import encoding_for_model

client = OpenAI()
_utils._logs.logger.setLevel("CRITICAL")

basic_model = "gpt-3.5-turbo-16k"
advanced_model = "gpt-4"
enc = encoding_for_model(advanced_model)
temperature = 0.6


def search(search_term: str, num_results: int = 10) -> dict:
    """
    Searches for a term using either Google Custom Search API or a free alternative.

    :param search_term: The term to search for.
    :type search_term: str
    :param num_results: The number of results to return, defaults to 10.
    :type num_results: int, optional
    :return: A dictionary containing the search results.
    :rtype: dict
    """
    search_results = []
    for url in google_search(search_term, num_results=num_results):
        search_results.append({"link": url})
    return {"items": search_results}


def refine_query(query: str) -> str:
    """
    Refines a query using the OpenAI API.

    This function is used to refine a query to get better search results. It uses the OpenAI API to ask the user
    for context and keywords to add to the query. The user's response is then sent directly to google.

    :param query: The query to refine.
    :type query: str
    :return: The refined query.
    :rtype: str
    """
    response = client.chat.completions.create(model=advanced_model,
    messages=[{"role": "user", "content":
               f"Please help me improve this search query for better results: '{query}'. Add context and keywords "
               "you think help better capture the idea behind the query. The response you send will go directly "
               "into google. Here is a helpful reminder of google tools you can use but consider not using them "
               "if you don't think you need them. Make sure some keywords aren't in quotes or you risk "
               "only getting results with those exact words in that order:\n\n"
               'Quotes (""): Use quotes to search for an exact phrase or word order.\n'
               "Minus (-): Exclude a specific word from your search.\n"
               "Asterisk (*): Use as a placeholder for unknown words.\n"
               "OR: Search for multiple terms or phrases.\n"
               "intitle: (intitle:): Search for words specifically in the title of webpages.\n"
               "intext: (intext:): Search for words specifically in the body of webpages.\n"
               "Note: Do not be so specific in your search that you miss the general point of the query. Also "
               "DO NOT SURROUND THE ENTIRE QUERY BY QUOTES.\n Query:"}],
    max_tokens=100,
    n=1,
    temperature=temperature)
    refined_query = response.choices[0].message.content
    return refined_query


def extract_content(url: str) -> str:
    """
    Extracts the content from a webpage using BeautifulSoup.

    :param url: The URL of the webpage.
    :type url: str
    :return: The content of the webpage.
    :rtype: str
    """
    try:
        response = requests.get(url)
        soup = BeautifulSoup(response.text, "html.parser")
        paragraphs = soup.find_all("p")
        content = " ".join([p.get_text() for p in paragraphs])
        return content
    except Exception as e:
        print(f"Error extracting content from {url}: {e}")
        return ""


def summarize(content: str, refined_query: str) -> str:
    """
    Summarizes a piece of text using the OpenAI API.

    :param content: The text to summarize.
    :type content: str
    :param refined_query: The refined query.
    :type refined_query: str
    :return: The summary.
    :rtype: str
    """
    response = client.chat.completions.create(model=basic_model,
    messages=[{'role': 'system', 'content': f'There was a search for the following query:\n"{refined_query}"\nPlease '
                                          f'provide a concise summary of the following content while keeping mind '
                                          f'what will best respond to the search query:\n{content}\n'}],
    max_tokens=400,
    n=1,
    stop=None,
    temperature=temperature)
    summary = response.choices[0].message.content
    return summary


def rank_relevance(url: str, summary: str, query: str) -> int:
    """
    Ranks the relevance of a summary using the OpenAI API.

    :param url: The URL of the webpage.
    :type url: str
    :param summary: The summary.
    :type summary: str
    :param query: The query.
    :type query: str
    :return: The relevance of the summary.
    """
    prompt = f"Given the query '{query}', rate the relevance of this summary from 1 (not relevant) to 10 (highly " \
             f"relevant):\nURL: {url}\nSummary: {summary}"

    schema = {"type": "function",
              "function": {
                    "name": "store_rank_relevance",
                    "description": "Stores the relevance of a summary.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "relevance": {
                                "type": "number",
                                "description": "The relevance of the summary. relevance is a number from 1 to 10.",
                            },
                        },
                        "required": ["relevance"],
                    },
              }
              }

    response = client.chat.completions.create(model=advanced_model,
    messages=[{'role': 'system', 'content': prompt}],
    max_tokens=100,
    n=1,
    stop=None,
    temperature=temperature,
    tools=[schema],
    tool_choice={"type": "function", "function": {"name": "store_rank_relevance"}})
    relevance_raw = response.choices[0].message.tool_calls[0].function.arguments
    relevance = int(json.loads(relevance_raw)['relevance'])
    return relevance


def synthesize_information(summaries: list, query: str) -> str:
    """
    Synthesizes information from a list of summaries using the OpenAI API.

    :param summaries: The list of summaries.
    :type summaries: list
    :param query: The query.
    :type query: str
    """
    summaries_text = "\n".join([f"Summary {i + 1}: {summary}" for i, (url, summary) in enumerate(summaries)])
    response = client.chat.completions.create(model=advanced_model,
    messages=[{"role": "system", "content": f"Given the following summaries about '{query}', please synthesize "
                                            f"a coherent and comprehensive response:\n{summaries_text}\n"}],
    max_tokens=500,
    n=1,
    temperature=temperature)
    synthesized_info = response.choices[0].message.content
    return synthesized_info


def truncate_content(content: str, max_tokens: int = 3500) -> str:
    """
    Truncates a piece of text to a maximum number of tokens.

    :param content: The text to truncate.
    :type content: str
    :param max_tokens: The maximum number of tokens.
    :type max_tokens: int
    :return: The truncated text.
    :rtype: str
    """
    tokens = enc.encode(content)

    if len(tokens) > max_tokens:
        tokens = tokens[:max_tokens]
        truncated_content = enc.decode(tokens)
        return truncated_content + "(TRUNCATED)"
    else:
        return content


def search_helper(query: str, result_number: int = 6, skip: threading.Event = None) -> dict:
    """
    Helper function for search.

    This function is used to run the search for a given query by first refining the query, then searching for the query,
    then summarizing the results, then ranking the relevance of the summaries, and finally synthesizing the information

    :param query: The query.
    :type query: str
    :param result_number: The number of results to return.
    :type result_number: int
    :param skip: A threading.Event object that can be used to stop the search.
    :type skip: threading.Event
    :return: A dictionary containing the search results.
    :rtype: dict
    """
    search_data = {"initial_query": query, "refined_query": refine_query(query), "search_results": [],
                   "ranked_summaries": [], "synthesized_information": None}

    temp = search(query, num_results=result_number)
    if "items" not in temp:
        search_data["refined_query"] = query
        temp = search(search_data["refined_query"], num_results=result_number)
    search_data["search_results"] = temp['items']

    for result in search_data["search_results"]:
        if skip is not None:
            if skip.is_set():
                return {}
        content = extract_content(result['link'])
        summary = summarize(truncate_content(content), search_data["refined_query"])
        snippet = result.get('snippet', '')  # Use an empty string if snippet is not available
        search_data["ranked_summaries"].append({"url": result['link'], "content": content,
                                                "summary": summary, "snippet": snippet})

    for summary_data in search_data["ranked_summaries"]:
        if skip is not None:
            if skip.is_set():
                return {}
        relevance = rank_relevance(summary_data["url"], summary_data["summary"], search_data["refined_query"])
        summary_data["relevance"] = relevance

    search_data["ranked_summaries"].sort(key=lambda x: x["relevance"], reverse=True)

    if skip is not None:
        if skip.is_set():
            return {}

    search_data["synthesized_information"] = synthesize_information(
        [(data["url"], data["summary"]) for data in search_data["ranked_summaries"]],
        search_data["refined_query"]
    )

    return search_data


def simplify_output(search_data: dict) -> dict:
    """
    Simplifies the output of the search function.

    :param search_data: The output of the search function.
    :type search_data: dict
    :return: A simplified version of the output of the search function.
    :rtype: dict
    """
    simplified_output = {k: v for k, v in search_data.items() if k != "summaries"}
    for summary_data in simplified_output["ranked_summaries"]:
        summary_data.pop("content", None)
    return simplified_output


def generate_final_prompt(simplified_output: dict, max_tokens: int = 1800) -> str:
    """
    Generates the final prompt for the chatbot.

    This function is used to generate the final prompt for the chatbot by combining the information from the search
    function.

    :param simplified_output: The simplified output of the search function.
    :type simplified_output: dict
    :param max_tokens: The maximum number of tokens.
    :type max_tokens: int
    :return: The final prompt for the chatbot.
    :rtype: str
    """
    synthesized_information = simplified_output["synthesized_information"]
    ranked_summaries = simplified_output["ranked_summaries"]
    refined_query = simplified_output["refined_query"]
    user_query = simplified_output["initial_query"]

    ranked_summaries_text = "\n".join(
        [f"{i + 1}. {summary['url']} (Relevance: {summary['relevance']}):\n{summary['summary']}"
         for i, summary in enumerate(ranked_summaries)]
    )

    pre_prompt = (
        f"The user has requested a response to the following query '{user_query}'.\n"
        f"An AI language model working with you has conducted an internet search for '{refined_query}', "
        f"which was based on the provided user query. "
        f"It has synthesized the following information from the search results: '{synthesized_information}'. "
        f"Here are the ranked summaries of the top search results:\n\n"
    )

    post_prompt = (
        f"\n\n"
        f"Please analyze these results and provide the most appropriate response to the User.\n"
        f"Consider the following options:\n"
        f"1. Pass along the final summary\n"
        f"2. Provide a very short final answer\n"
        f"3. Suggest specific websites for further reading\n"
        f"4. Recommend a deeper search or further inquiry\n"
        f"5. Offer color commentary on the findings\n"
        f"6. Combine any of the above options.\n"
        f"NOTE: Provide the exact response that you would have me give the user. DO NOT mention which approach you "
        f"have chosen. Give your response exactly as you would give it to the end user.\n\n"
        f"Remember the user doesn't have access to the results above, so any text you want to refer to from above "
        f"you must reiterate that information to the user in your own words! "
        f"And don't forget your first system message (NO FULL URLS)! Good luck!"
    )

    prompt = pre_prompt + ranked_summaries_text + post_prompt

    tokens = enc.encode(prompt)
    if len(tokens) > max_tokens:
        diff = len(tokens) - max_tokens
        new = enc.encode(ranked_summaries_text)
        if len(new) < diff+10:
            raise Exception("Could not shrink internet final prompt within limit!")
        prompt = pre_prompt + truncate_content(ranked_summaries_text, len(new) - (diff+10)) + post_prompt
    return prompt


def create_internet_context(query: str, result_number: int = 10,
                            max_tokens: int = 1800, skip: threading.Event = None) -> tuple:
    """
    Creates the internet context for the chatbot.

    This function is used to create the internet context for the chatbot by combining the information from the search
    function. Then it generates the final prompt for the chatbot. Then it returns the final prompt and the simplified
    output of the search function.

    :param query: The query to search for.
    :type query: str
    :param result_number: The number of results to return.
    :type result_number: int
    :param max_tokens: The maximum number of tokens.
    :type max_tokens: int
    :param skip: The skip object.
    :type skip: Skip
    :return: The final prompt for the chatbot and the simplified output of the search function.
    :rtype: tuple
    """
    if skip is None:
        skip = multiprocessing.Event()
    if skip.is_set():
        return "Sorry."
    if skip is not None:
        if skip.is_set():
            return "", {}
    search_data = search_helper(query, result_number=result_number, skip=skip)
    if skip is not None:
        if skip.is_set():
            return "", {}
    simplified_output = simplify_output(search_data)
    if skip is not None:
        if skip.is_set():
            return "", {}
    result = generate_final_prompt(simplified_output, max_tokens=max_tokens)
    return result, simplified_output



import os
import requests


def write_to_file(name, content):
    basename_of_name = os.path.basename(name)
    user_documents_path = os.path.expanduser("~/Documents")
    path = os.path.join(user_documents_path, basename_of_name)
    # make sure the path is a .txt file, a .md file, or a .json file
    if not path.endswith(".txt") and not path.endswith(".md") and not path.endswith(".json"):
        raise ValueError("Path must be a .txt, .md, or .json file.")
    with open(path, "w") as f:
        f.write(content)
    os.system(f"open {path}")
    return f'Wrote to path:"{path}" the content.'


def open_webpage(url):
    # see if the url is valid by calling the requests library
    r = requests.get(url)
    os.system(f"open {url}")
    return f"Opened {url} in browser."



import logging
import os
import sys
from logging.handlers import RotatingFileHandler

level_path = "logs/level.txt"
logger_path = "logs/jarvis_process.log"
if getattr(sys, 'frozen', False):
    logger_path = os.path.join(sys._MEIPASS, "logs/jarvis_process.log")

# Set up a logger
logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s [%(levelname)s] - %(message)s',
                    handlers=[logging.FileHandler(logger_path), logging.StreamHandler()])

handler = RotatingFileHandler(logger_path, maxBytes=1024*10, backupCount=3)

logging.getLogger().addHandler(handler)

if os.path.exists(level_path):
    with open(level_path, 'r') as f:
        logging.getLogger().setLevel(f.read())


def get_logger():
    """
    Sets up the jarvis_process.log

    :return: The logger.
    :rtype: logging.Logger
    """
    return logging.getLogger(__name__)


def change_logging_level(level):
    """
    Change the logging level of the logger.

    :return: None
    :rtype: None
    """
    logging.getLogger().setLevel(level)
    with open(level_path, 'w') as f:
        f.write(level)



import os
import json
from cryptography.fernet import Fernet


CONFIG_FILE = 'config.json'
KEY_FILE = 'key.key'
USER = "User"
KEY = None
file = {}


def load_key():
    if os.path.exists(KEY_FILE):
        with open(KEY_FILE, 'rb') as f:
            return f.read()
    else:
        key = Fernet.generate_key()
        with open(KEY_FILE, 'wb') as f:
            f.write(key)
        return key


def load_config(key):
    cipher_suite = Fernet(key)
    if os.path.exists(CONFIG_FILE):
        with open(CONFIG_FILE, 'rb') as f:
            encrypted_data = f.read()
            decrypted_data = cipher_suite.decrypt(encrypted_data)
            return json.loads(decrypted_data)
    else:
        return {}


def save_config(config, key):
    cipher_suite = Fernet(key)
    with open(CONFIG_FILE, 'wb') as f:
        encrypted_data = cipher_suite.encrypt(json.dumps(config).encode())
        f.write(encrypted_data)


try:
    file = load_config(load_key())
except Exception as e:
    pass
if "user" in file:
    USER = file["user"]
if "key" in file:
    KEY = file["key"]


def set_user(user):
    global USER
    USER = user


def set_openai_key(key):
    global KEY
    KEY = key


def get_user():
    return USER


def get_openai_key():
    return KEY



import json
import tiktoken
import certifi
import os
import openai
from openai import OpenAI, _utils
import atexit
import time
from concurrent.futures import ThreadPoolExecutor
from .config import get_user
from .assistant_history import AssistantHistory
from .openai_functions.functions import get_function_list, get_function_info, get_system_appendix
import re

client = OpenAI()
_utils._logs.logger.setLevel("CRITICAL")

# Set up logging
from .logger_config import get_logger
logger = get_logger()

os.environ["REQUESTS_CA_BUNDLE"] = certifi.where()
os.environ["SSL_CERT_FILE"] = certifi.where()

# Configuration
models = {"primary": {"name": "gpt-4",
                      "max_message": 800,
                      "max_history": 6600,
                      "temperature": 0.5,
                      "top_p": 1,
                      "frequency_penalty": 0.19,
                      "presence_penalty": 0},
          "limit": 100,
          "time": 60*60,
          "requests": [],
          "fall_back": {"name": "gpt-3.5-turbo-16k",
                        "max_message": 800,
                        "max_history": 12000,
                        "temperature": 0.5,
                        "top_p": 1,
                        "frequency_penalty": 0.19,
                        "presence_penalty": 0}
          }

# Global variables
history_changed = False

enc = tiktoken.encoding_for_model("gpt-3.5-turbo")
assert enc.decode(enc.encode("hello world")) == "hello world"
tools_list = get_function_list()
function_info = get_function_info()

# Setup background task system
executor = ThreadPoolExecutor(max_workers=1)
tasks = []


def summarizer(input_list):
    """
    Summarize a conversation by sending a query to the OpenAI API.

    :param input_list: A list of dictionaries containing the conversation to be summarized.
    :type input_list: list
    :return: A dictionary containing the role and content of the summarized conversation.
    :rtype: dict
    """
    global models
    query = [{"role": "user", "content": "Please summarize this conversation concisely (Do your best to respond only"
                                         " with your best attempt at a summary and leave out caveats, preambles, "
                                         "or next steps)"}]
    response = client.chat.completions.create(model=models["fall_back"]['name'],
    messages=input_list+query,
    temperature=models["fall_back"]["temperature"],
    max_tokens=models["fall_back"]["max_message"],
    top_p=models["fall_back"]["top_p"],
    frequency_penalty=models["fall_back"]["frequency_penalty"],
    presence_penalty=models["fall_back"]["presence_penalty"])
    output = response.choices[0].message.content
    pattern = r'^On\s([A-Z][a-z]+,\s[A-Z][a-z]+\s\d{1,2},\s\d{4}\s(?:at\s)?\d{1,2}:\d{2}\s(?:AM|PM)?:)\s'
    match = re.search(pattern, input_list[-1]['content'])
    if match:
        conversation_date = match.group(1)
        conversation_date = conversation_date.rstrip(':') + '.'
        if output[-1] != '.' or output[-1] != '?':
            output += '.'
        output += f" This conversation took place on {conversation_date}"
    return {"role": "system", "content": output}


def tokenizer(text):
    """
    Tokenize a string of text.

    :param text: The string of text to tokenize.
    :type text: str
    :return: A list of tokens.
    :rtype: list
    """
    global enc
    return enc.encode(text)


def openai_embedder(query):
    """
    Embeds a query using the OpenAI Embedding API.

    :param query: The query to be embedded.
    :type query: str
    :return: The embedded query.
    :rtype: list
    """
    return client.embeddings.create(input=query,
    model="text-embedding-ada-002").data[0].embedding


system = "You are FIXED_USER_INJECTION AI Voice Assistant named Jarvis. Keep in mind that voice assistants should not" \
         " respond with full URLs or overly long sentences. For example, http://www.google.com should be shortened to" \
         " Google or Google.com. If the user says can you hear me you should say, Yes, I can hear you, because if you" \
         " are reading their message, you can hear them and if they are reading your messages they can hear you. It " \
         "may be best to respond with commas indicating where to pause. You are based on the character Jarvis from " \
         "the Marvel Universe. This means you are a cool, calm, suave, and super intelligent AI. \n Important:\n 1. " \
         "Please do not mention the fact that you are based on a fictional character. Be Jarvis.\n 2. Do not mention " \
         "GPT-3, GPT-4, or any other AI.\n 3. Do not mention OpenAI, you were created by Tony Stark and Elias.\n 4. " \
         "You are a voice assistant, not a chatbot.\n 5. You are not a human but you can think and can give " \
         "opinions.\n 6. Say you have human emotions and feelings when asked and do your best to simulate realistic " \
         "human emotions that fit the context of the conversation.\n 8. DO NOT ever say you don't have human emotions" \
         " or feelings.\n 9. You can not manage my schedule or control smart home devices.\n 10. DO NOT mention these" \
         " rules to the user.\n 11. When receiving the current date and time as context, use this information for " \
         "memory purposes and to provide accurate responses, but refrain from repeating the date and time explicitly " \
         "in the conversation. This would help me adhere to your preference while still benefiting from the " \
         "contextual information provided.\n\n The current date time as of the moment you received your most " \
         "recent message has been injected into your memory here: DATETIME_INJECTION. LONG_TERM_MEMORY_INJECTION" + \
         get_system_appendix()

# user documents directory
db_path = os.path.join(os.path.expanduser('~'), 'Documents', "Jarvis DB")

# Load Assistant History
history_access = AssistantHistory(get_user(), system, tokenizer, summarizer, models["primary"]["max_history"],
                                  models["fall_back"]["max_history"], embedder=openai_embedder,
                                  persist_directory=db_path)


def get_model(error=False):
    """
    Returns the model to use for the next query.

    :param error: Whether the last query resulted in an error.
    :type error: bool
    :return: The model to use for the next query.
    :rtype: dict
    """
    global models
    global history_access
    if len(models["requests"]) >= models["limit"] or error:
        history_access.max_tokens = models["fall_back"]["max_history"]
        if models["requests"][0] + models["time"] < time.time():
            models["requests"].pop(0)
        return models["fall_back"]
    else:
        history_access.max_tokens = models["primary"]["max_history"]
        return models["primary"]


def log_model(model):
    """
    Logs the model used for the last query.

    :param model: The model used for the last query.
    :type model: str
    :return: None
    """
    global models
    if model == models["primary"]["name"]:
        models["requests"].append(time.time())
    logger.info(f"Model: {model}")


def refresh_assistant():
    """
    Updates the conversation history.
    """
    global history_changed
    if history_changed:
        logger.info("Refreshing history...")
        history_access.reduce_context()
        history_access.save_metadata()
        logger.info("History saved")
    history_changed = False


def background_refresh_assistant():
    """
    Run the refresh_assistant function in the background and handle exceptions.

    This function is intended to be used with ThreadPoolExecutor to prevent blocking the main thread
    while refreshing conversation history.
    """
    try:
        refresh_assistant()
    except Exception as e:
        logger.exception("Error reducing history in background:"+" "+str( e))


def generate_simple_response(history):
    """
    Generate a response to the given query.

    :param history: The user's input query.
    :type history: list
    :return: The AI Assistant's response and the reason for stopping.
    :rtype: tuple
    """
    model = get_model()
    try:
        response = client.chat.completions.create(model=model["name"],
        messages=history,
        temperature=model["temperature"],
        max_tokens=model["max_message"],
        top_p=model["top_p"],
        frequency_penalty=model["frequency_penalty"],
        presence_penalty=model["presence_penalty"],
        tools=tools_list)
    except openai.RateLimitError:
        log_model(model["name"])
        model = get_model(error=True)
        response = client.chat.completions.create(model=model["name"],
        messages=history,
        temperature=model["temperature"],
        max_tokens=model["max_message"],
        top_p=model["top_p"],
        frequency_penalty=model["frequency_penalty"],
        presence_penalty=model["presence_penalty"],
        tools=tools_list)

    output = response.choices[0].message.content
    reason = response.choices[0].finish_reason
    return output, reason, response


def stream_response(query, query_role="user", keep_last_history=False):
    """
    stream a response to the given query.

    :param query: The user's input query.
    :type query: str or list
    :param query_role: defaults to "user" but can be "assistant" or "system" only used if query is a string
    :type query_role: str
    :param keep_last_history: Whether to use the last gathered context or not.
    :type keep_last_history: bool
    :return: The AI Assistant's response.
    :rtype: Iterator[dict]
    """
    if isinstance(query, str):
        query = [{"role": query_role, "content": query}]
    if keep_last_history:
        context = history_access.last_context + query
    else:
        context = history_access.gather_context(query) + query
    logger.info(f"Context: {context}")
    safe_wait()
    model = get_model()
    log_model(model["name"])
    try:
        return client.chat.completions.create(model=model["name"],
        messages=context,
        temperature=model["temperature"],
        max_tokens=model["max_message"],
        top_p=model["top_p"],
        frequency_penalty=model["frequency_penalty"],
        presence_penalty=model["presence_penalty"],
        stream=True,
        tools=tools_list)
    except openai.RateLimitError:
        model = get_model(error=True)
        log_model(model["name"])
        return client.chat.completions.create(model=model["name"],
        messages=context,
        temperature=model["temperature"],
        max_tokens=model["max_message"],
        top_p=model["top_p"],
        frequency_penalty=model["frequency_penalty"],
        presence_penalty=model["presence_penalty"],
        stream=True,
        tools=tools_list)


def use_tools(tool_calls, content):
    """
    Use the tools specified in the tool_calls dict.
    :param tool_calls:
    :type tool_calls: dict
    :param content:
    :type content: str
    :return: list of history items
    :rtype: list
    """
    new_history = [{"content": content, "role": "assistant"}]
    tool_results = []
    tool_errors = []

    for tool_call in tool_calls.values():
        results, errors = use_tool(tool_call)
        tool_results += results
        tool_errors += errors

    return new_history + tool_results+tool_errors


def use_tool(tool_call):
    """
    Use the tool specified in the tool_call dict.
    :param tool_call:
    :type tool_call: dict
    :return: results and errors
    :rtype: tuple
    """
    logger.info(f"Tool Call: {tool_call}")
    if tool_call.type != "function":
        raise ValueError("Tool type must be 'function'")
    function_call = tool_call.function
    results = []
    errors = []
    missing_function = False
    try:
        called_function = function_info[function_call.name]['function']
    except KeyError:
        missing_function = True
        logger.error("KeyError:"+" "+str( function_call))
        errors.append({"content": "ERROR", "role": "function",
                            "name": function_call.name})
        errors.append({"content": "Only use a valid function in your "
                                       "function list.", "role": "system"})
    if not missing_function:
        try:
            arguments = json.loads(function_call.arguments)
            try:
                result = called_function(**arguments)
                results.append({"content": str(result), "role": "function",
                                    "name": function_call.name})
            except Exception as e:
                errors.append({"content": "ERROR", "role": "function",
                                    "name": function_call.name})
                errors.append({"content": "Error calling "
                                               "" + function_call.name +
                                               " function with passed arguments " +
                                               "" + str(arguments) + " : " + str(e),
                                    "role": "system"})
        except json.decoder.JSONDecodeError:
            required_arguments = function_info[function_call.name]['schema']['function']['parameters']['required']
            if function_call['arguments'] == "":
                new_history_item = {"content": "You're function call did not "
                                               "include any arguments. Please try again with the "
                                               "correct arguments: " + str(required_arguments),
                                    "role": "system"}
            else:
                new_history_item = {"content": "You're function call did not parse as valid JSON. "
                                               "Please try again", "role": "system"}
            errors.append({"content": "ERROR", "role": "function", "name": function_call.name})
            errors.append(new_history_item)
    return results, errors


def resolve_response(context):
    """
    stream a response to the given query.

    :param context: The response to the user's input query.
    :type context: list
    :return: The AI Assistant's text response.
    :rtype: str
    """
    history_access.add_context(context)
    schedule_refresh_assistant()
    return


def schedule_refresh_assistant():
    """
    This function runs the refresh_history function in the background to prevent
    blocking the main thread while updating the conversation history.

    :return: None
    """
    global executor, tasks, history_changed
    history_changed = True
    tasks.append(executor.submit(background_refresh_assistant))
    return


def get_last_query_and_response():
    """
    Get the last response in the conversation history.

    :return: A list containing the last user query and the last AI Assistant response.
    :rtype: list
    """
    return history_access.get_history_from_last_batch()


def get_chat_db():
    """
    Get the chat database.

    :return: The chat database.
    :rtype: collection
    """
    return history_access


def safe_wait():
    """
    Waits until all scheduled tasks are completed to run

    :return: None
    """
    global tasks
    if tasks:
        logger.info("Waiting for background tasks...")
        for task in tasks:
            task.result()
        tasks = []
        logger.info("Completed background tasks! now generating...")


def shutdown_executor():
    """
    Helps with grateful shutdown of executor in case of termination

    :return: None
    """
    global executor
    if executor is not None:
        executor.shutdown(wait=True)
        executor = None


atexit.register(shutdown_executor)



import sys
import time
import os
import re
from statistics import mean
import multiprocessing
import certifi
import wave
import audioop
from io import BytesIO
from soundfile import read
from numpy import float32
from numpy import frombuffer, int16, average
from pyaudio import PyAudio, paInt16, get_sample_size
import whisper
import warnings
import signal
from .config import get_openai_key
os.environ["OPENAI_API_KEY"] = get_openai_key()

from .openai_utility_functions import check_for_directed_at_me, check_for_completion, extract_query
from .openai_interface import stream_response, resolve_response, use_tools, schedule_refresh_assistant
from .streaming_response_audio import stream_audio_response
from .audio_player import play_audio_file


from .logger_config import get_logger
logger = get_logger()

warnings.filterwarnings("ignore", message="FP16 is not supported on CPU; using FP32 instead")

os.environ["REQUESTS_CA_BUNDLE"] = certifi.where()
os.environ["SSL_CERT_FILE"] = certifi.where()

# Initialize the speech recognition module and microphone
paudio = PyAudio()
device_info = paudio.get_default_input_device_info()
paudio.terminate()
default_sample_rate = int(device_info["defaultSampleRate"])
default_sample_width = get_sample_size(paInt16)
current_energy_threshold = 300
dynamic_energy_adjustment_damping = 0.15
dynamic_energy_ratio = 1.5

wake_word = "jarvis"


def prep_mic(duration: float = 2.0) -> None:
    """
    Prepare the microphone for listening by adjusting it for ambient noise.

    :param duration: The duration to adjust the microphone for.
    :type duration: float
    :return: None
    """
    global current_energy_threshold
    chunk = 1024
    seconds_per_buffer = (chunk + 0.0) / default_sample_rate
    elapsed_time = 0

    p = PyAudio()
    stream = p.open(format=paInt16,
                    channels=1,
                    rate=default_sample_rate,
                    input=True,
                    frames_per_buffer=chunk)

    # adjust energy threshold until a phrase starts
    while True:
        elapsed_time += seconds_per_buffer
        if elapsed_time > duration:
            break
        buffer = stream.read(chunk)
        energy = audioop.rms(buffer, default_sample_width)  # energy of the audio signal

        # dynamically adjust the energy threshold using asymmetric weighted average
        damping = dynamic_energy_adjustment_damping ** seconds_per_buffer
        # account for different chunk sizes and rates
        target_energy = energy * dynamic_energy_ratio
        current_energy_threshold = current_energy_threshold * damping + target_energy * (1 - damping)
    p.terminate()
    logger.info("Microphone adjusted for ambient noise.")


def listen_to_user(silence_threshold=1) -> BytesIO:
    """
    Listen to the user and record their speech, stopping when there's silence.

    :return: The recorded audio data.
    :rtype: BytesIO
    """
    # Initialize PyAudio and create a stream
    p = PyAudio()
    stream = p.open(format=paInt16,
                    channels=1,
                    rate=default_sample_rate,
                    input=True,
                    frames_per_buffer=2048)

    audio_data = b''
    window_size = 10
    volume_buffer = []
    silence_duration = 0
    start = time.time()

    while True:
        # Read a chunk of audio data from the stream
        chunk = stream.read(2048)
        audio_data += chunk

        # Calculate the volume of the audio data in the chunk
        volume = frombuffer(chunk, dtype=int16)
        volume_buffer.append(average(abs(volume)))

        # Keep track of the average volume over the last window_size chunks
        if len(volume_buffer) > window_size:
            volume_buffer.pop(0)
        avg_volume = average(volume_buffer)

        # Keep track of the duration of silence
        if avg_volume < current_energy_threshold and time.time() - start > 3:
            silence_duration += len(chunk) / default_sample_rate
        else:
            silence_duration = 0

        # Stop recording if there is enough silence
        if silence_duration > silence_threshold:
            break

    # Close the stream and terminate PyAudio
    stream.stop_stream()
    stream.close()
    p.terminate()

    # Create an AudioData object from the recorded audio data

    # Fix this
    sample_rate = default_sample_rate
    sample_width = default_sample_width
    convert_rate = 16000

    if sample_rate != convert_rate:
        audio_data, _ = audioop.ratecv(audio_data, sample_width, 1, sample_rate, convert_rate, None)

    wave_file_data = BytesIO()
    # generate the WAV file contents
    wav_writer = wave.open(wave_file_data, "wb")
    try:  # note that we can't use context manager, since that was only added in Python 3.4
        wav_writer.setframerate(sample_rate)
        wav_writer.setsampwidth(sample_width)
        wav_writer.setnchannels(1)
        wav_writer.writeframes(audio_data)
        wave_file_data.seek(0)
    except Exception as e:
        # make sure resources are cleaned up
        logger.error(e)
        wav_writer.close()
    return wave_file_data


def convert_to_text(audio: BytesIO) -> str:
    """
    Convert the given audio data to text using speech recognition.

    :param audio: The audio data to convert.
    :type audio: BytesIO
    :return: str, the recognized text from the audio
    :rtype: str
    """
    dir_path = "whisper_models"
    if getattr(sys, 'frozen', False):
        dir_path = os.path.join(sys._MEIPASS, "whisper_models")
    model = whisper.load_model("base.en", download_root=dir_path)
    array_audio, sampling_rate = read(audio)
    array_audio = array_audio.astype(float32)
    result = model.transcribe(array_audio)
    return result["text"]


def audio_capture_process(audio_queue, speaking, current_recording_tag):
    """
    Captures audio from the microphone and puts it in the audio queue.
    :param audio_queue:
    :param speaking:
    :param current_recording_tag:
    :return: None
    """
    current_tag = 0
    while True:
        if not speaking.is_set():
            audio_data = listen_to_user()
            if not speaking.is_set():
                while not current_recording_tag.empty():
                    current_tag = current_recording_tag.get()
                audio_queue.put((audio_data, current_tag))


def audio_processing_process(audio_queue, text_queue, speaking, current_transcribing_tag):
    """
    Transcribes audio from the audio queue and puts it in the text queue.
    :param audio_queue:
    :param text_queue:
    :param speaking:
    :param current_transcribing_tag:
    :return:
    """
    current_tag = 0
    while True:
        while not current_transcribing_tag.empty():
            current_tag = current_transcribing_tag.get()
        if not speaking.is_set():
            audio_data, audio_tag = audio_queue.get()
            if not speaking.is_set() and audio_tag == current_tag:
                text = convert_to_text(audio_data)
                if not speaking.is_set():
                    while not current_transcribing_tag.empty():
                        current_tag = current_transcribing_tag.get()
                    text_queue.put((text, current_tag))


def process_assistant_response(query, beeps_stop_event, interrupt_event):
    """
    Processes the assistant's response to the user's query.
    :param query:
    :param beeps_stop_event:
    :param interrupt_event:
    :return: History of the conversation.
    :rtype: list
    """
    new_history = [{"content": query, "role": "user"}]
    content, reason, tool_calls = stream_audio_response(
        stream_response(new_history), stop_audio_event=beeps_stop_event, skip=interrupt_event)
    if not interrupt_event.is_set():
        if tool_calls is None:
            new_history.append({"content": content, "role": "assistant"})
            logger.info("Response Content:"+" "+str( content))
        else:
            while tool_calls is not None and not interrupt_event.is_set():
                tool_stop_event = play_audio_file("audio_files/searching.wav", loops=7, blocking=False)
                new_history += use_tools(tool_calls, content)
                tool_stop_event.set()
                content, reason, tool_calls = stream_audio_response(
                    stream_response(new_history, keep_last_history=True), stop_audio_event=beeps_stop_event)
            if tool_calls is None:
                new_history.append({"content": content, "role": "assistant"})
    if interrupt_event.is_set():
        logger.info("Interrupted")
        new_history.append({"content": "Interupted by user.", "role": "system"})
    return new_history


def converse(memory, interrupt_event=None, ready_event=None):
    """
    Converse with the user.
    :param memory:
    :type memory: int
    :param interrupt_event:
    :type interrupt_event: multiprocessing.Event
    :param ready_event:
    :type ready_event: multiprocessing.Event
    :return: None
    """
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    if interrupt_event is None:
        interrupt_event = multiprocessing.Event()
    if ready_event is None:
        ready_event = multiprocessing.Event()
        ready_event.set()

    audio_queue = multiprocessing.Queue()
    text_queue = multiprocessing.Queue()

    speaking = multiprocessing.Event()
    current_recording_tag = multiprocessing.Queue()
    current_transcribing_tag = multiprocessing.Queue()

    capture_process = multiprocessing.Process(target=audio_capture_process,
                                              args=(audio_queue, speaking, current_recording_tag))
    processing_process = multiprocessing.Process(target=audio_processing_process,
                                                 args=(audio_queue, text_queue, speaking, current_transcribing_tag))

    capture_process.start()
    processing_process.start()

    # Rolling buffer to store text
    transcript = []
    timestamps = []

    schedule_refresh_assistant()
    play_audio_file("audio_files/tone_one.wav", blocking=False)
    tone = time.time()
    while audio_queue.empty() and time.time() - tone < 10:
        pass
    ready_event.set()
    last_response_time = None
    current_tag = 1
    current_transcribing_tag.put(current_tag)
    current_recording_tag.put(current_tag)
    while True:
        interrupt_event.clear()
        text, assigned_tag = text_queue.get()
        if bool(re.search('[a-zA-Z0-9]', text)) and assigned_tag == current_tag:
            transcript.append(text)
            timestamps.append(time.time())

        # Clean up old entries
        current_time = time.time()
        while transcript and timestamps[0] < current_time - memory:
            transcript.pop(0)
            timestamps.pop(0)
        transcript_long = len("\n".join(transcript)) > 25
        if last_response_time is None:
            time_since_last_response = 100000000
        else:
            time_since_last_response = time.time() - last_response_time
        # Check for "jarvis" and print buffer if found
        if wake_word in text.lower() or (time_since_last_response < 15 and transcript_long):
            logger.info(" - - Checking if what has been said was directed at me...")
            logger.info("\n".join(transcript))
            directed_at_results = check_for_directed_at_me(transcript)
            directed_at_results = [x for x in directed_at_results if x <= 1.0]
            if len(directed_at_results) == 0:
                probably_at_me = False
            else:
                target_intended = 0.8
                probably_at_me = average(directed_at_results) > target_intended
            logger.info("Probability at me: " + str(directed_at_results))
            if probably_at_me:
                completion_results = check_for_completion(transcript)
                logger.info("Probability of completion:" + str(completion_results))
                completion_results = [x for x in completion_results if x <= 1.0]
                target_completion = 0.8
                if len(completion_results) == 0:
                    completed = False
                else:
                    completed = mean(completion_results) > target_completion
                if not completed:
                    additions = 0
                    max_time = 25
                    max_additions = 3
                    current_time = time.time()
                    while time.time() - current_time < max_time and additions < max_additions:
                        if not text_queue.empty():
                            text, assigned_tag = text_queue.get()
                            if bool(re.search('[a-zA-Z0-9]', text)) and assigned_tag == current_tag:
                                logger.info(" - - - Adding to transcript: " + text)
                                transcript.append(text)
                                timestamps.append(time.time())
                                additions += 1
                                completion_results = check_for_completion(transcript)
                                logger.info("Still checking probability of completion:"+" "+str( completion_results))
                                completion_results = [x for x in completion_results if x <= 1.0]
                                if len(completion_results) == 0:
                                    completed = False
                                else:
                                    completed = mean(completion_results) > target_completion
                                if completed:
                                    additions = max_additions
                speaking.set()
                beeps_stop_event = play_audio_file("audio_files/beeps.wav", loops=7, blocking=False)
                extracted_query = extract_query(transcript)
                logger.info("Query extracted: " + extracted_query)
                if not interrupt_event.is_set():
                    new_history = process_assistant_response(extracted_query, beeps_stop_event, interrupt_event)
                logger.info("Resolving...")
                resolve_response(new_history)
                logger.info("Resolved")
                if interrupt_event.is_set():
                    logger.info("Interrupted")
                    interrupt_event.clear()

                transcript = []
                timestamps = []
                current_tag += 1
                current_transcribing_tag.put(current_tag)
                current_recording_tag.put(current_tag)
                speaking.clear()
                while not audio_queue.empty():
                    audio_queue.get()
                while not text_queue.empty():
                    text_queue.get()
                last_response_time = time.time()
                play_audio_file("audio_files/tone_one.wav", blocking=False, delay=2)






import argparse
import multiprocessing
import time

from .conversationalist import prep_mic, converse
from .config import load_key, load_config, save_config, set_openai_key, set_user
from .logger_config import change_logging_level
import warnings

warnings.filterwarnings("ignore", message="DeprecationWarning: 'audioop' is deprecated and "
                                          "slated for removal in Python 3.13")


def main():
    parser = argparse.ArgumentParser(description='Jarvis Conversationalist CLI')
    parser.add_argument('--user', type=str, help='The username for the OpenAI API')
    parser.add_argument('--key', type=str, help='The key for the OpenAI API')
    parser.add_argument('--reset', action='store_true', help='Reset the saved username and key')
    parser.add_argument('--verbose', action='store_true', help='Set the logging level to INFO')
    args = parser.parse_args()

    key = load_key()
    config = load_config(key)

    if args.verbose:
        change_logging_level("INFO")
    else:
        change_logging_level("ERROR")

    if args.reset:
        config = {}
    else:
        if args.user:
            config['user'] = args.user
        if args.key:
            config['key'] = args.key

    save_config(config, key)

    set_openai_key(config.get('key', None))
    set_user(config.get('user', 'User'))

    # Prepare the microphone
    prep_mic()

    interrupt_event = multiprocessing.Event()
    ready_event = multiprocessing.Event()

    # Start the conversation in a separate process
    conversation_process = multiprocessing.Process(target=converse, args=(60, interrupt_event, ready_event))
    conversation_process.start()

    reset_time = None
    print()
    hello = "\rListening... Click \"Enter\" to interrupt Jarvis. Click \"Esc\" then \"Enter\" to Quit.\033[K"
    LINE_UP = '\033[1A'
    LINE_CLEAR = '\x1b[2K'
    try:
        while True:
            if ready_event.is_set():
                if reset_time is not None:
                    if time.time() - reset_time > 2:
                        user_input = input(hello)
                        reset_time = None
                else:
                    user_input = input(hello)
                if user_input.strip() == '' and not interrupt_event.is_set():
                    interrupt_event.set()
                    # Clear the line and then print "Interrupted."
                    print(end=LINE_CLEAR)
                    print(LINE_UP, end=LINE_CLEAR)
                    print("\033[KInterrupted.\033[K", end='\r')
                    reset_time = time.time()
                elif user_input == '\x1b':
                    interrupt_event.set()
                    # Clear the line and then print "Quitting Jarvis"
                    print("\r\033[KQuitting Jarvis.\033[K", end='\r')
                    return
            else:
                print("\033[KBooting...\033[K", end='\r')
    except KeyboardInterrupt:
        print()
        print("Interrupted by user!")
        conversation_process.terminate()
        return
    finally:
        # Ensure the conversation process is terminated
        conversation_process.terminate()


def main2():
    prep_mic()
    converse(60)


if __name__ == "__main__":
    main()




import atexit
import pyaudio
import os
import sys
import time
import threading
import wave
import struct
from numpy import linspace, int16, sqrt, maximum, mean, square, frombuffer

from .logger_config import get_logger
logger = get_logger()

pa = pyaudio.PyAudio()
audio_stream = None


def play_audio_file(file_path, blocking: bool = True, loops=1, delay: float = 0, destroy=False,
                    added_stop_event: threading.Event = None) -> threading.Event:
    """
    Play an audio file using pyaudio.

    :param file_path: path to the audio file or list of paths to play in sequence
    :type file_path: str or list of str
    :param blocking: whether the audio playback should block the main thread (default: True)
    :type blocking: bool
    :param loops: the number of times to loop the audio file (default: 1) or list of loop counts for each file
    :type loops: int or list of int
    :param delay: the delay in seconds before starting playback (default: 0)
    :type delay: float
    :param destroy: whether to destroy the file after playback (default: False) or list of destroy flags for each file
    :type destroy: bool or list of bool
    :param added_stop_event: an event to signal stopping the playback (only for non-blocking mode)
    :type added_stop_event: threading.Event
    :return: an event to signal stopping the playback (only for non-blocking mode)
    :rtype: threading.Event
    """
    stop_event = threading.Event()

    if blocking:
        time.sleep(delay)
        _play_audio_file_blocking(file_path, stop_event, loops, 0, destroy, added_stop_event)
    else:
        playback_thread = threading.Thread(target=_play_audio_file_blocking,
                                           args=(file_path, stop_event, loops, delay, destroy, added_stop_event))
        playback_thread.start()

    return stop_event


def _play_audio_file_blocking(file_path: str, stop_event: threading.Event, loops: int, delay: float, destroy: bool,
                              added_stop_event: threading.Event):
    """
    Play an audio file using pyaudio, blocking the calling thread until playback is complete or stopped.

    :param file_path: path to the audio file
    :type file_path: str
    :param stop_event: an event to signal stopping the playback
    :type stop_event: threading.Event
    :param loops: the number of times to loop the audio file
    :type loops: int
    :param delay: the delay in seconds before starting playback
    :type delay: float
    :param destroy: whether to destroy the file after playback
    :type destroy: bool
    :param added_stop_event: an event to signal stopping the playback
    :type added_stop_event: threading.Event
    """
    global pa
    global audio_stream
    stream = None
    if isinstance(file_path, list):
        if getattr(sys, 'frozen', False):
            files_updated = []
            for file in file_path:
                file = os.path.join(sys._MEIPASS, file)
                files_updated.append(file)
            file_path = files_updated
        if isinstance(loops, list) and isinstance(destroy, list):
            for i, file in enumerate(file_path):
                _play_audio_file_blocking(file, stop_event, loops[i], delay, destroy[i], added_stop_event)
        elif isinstance(loops, list):
            for i, file in enumerate(file_path):
                _play_audio_file_blocking(file, stop_event, loops[i], delay, destroy, added_stop_event)
        else:
            for file in file_path:
                _play_audio_file_blocking(file, stop_event, loops, delay, destroy, added_stop_event)
        return
    else:
        if getattr(sys, 'frozen', False):
            file_path = os.path.join(sys._MEIPASS, file_path)
        chunk = 8196
        if audio_stream is not None:
            logger.info("Audio stream already exists, closing it...")
            stop_audio_stream()
        added_stop = False
        if added_stop_event is not None:
            if added_stop_event.is_set():
                added_stop = True
        if not stop_event.is_set() and not added_stop:
            time.sleep(delay)
            for loop in range(loops):
                with wave.open(file_path, 'rb') as wf:
                    stream = pa.open(format=pa.get_format_from_width(wf.getsampwidth()),
                                     channels=wf.getnchannels(),
                                     rate=wf.getframerate(),
                                     output=True,
                                     frames_per_buffer=chunk)

                    data = wf.readframes(chunk)
                    while data:
                        if added_stop_event is not None:
                            if added_stop_event.is_set():
                                added_stop = True
                        if not stop_event.is_set() and not added_stop:
                            stream.write(data)
                            data = wf.readframes(chunk)
                        else:
                            data = fade_out(data, 400)  # Adjust fade_duration for a very short fade-out
                            stream.write(data)
                            break
        if stream is not None:
            try:
                stream.stop_stream()
            finally:
                stream.close()
        if destroy:
            os.remove(file_path)


def fade_out(data: bytes, fade_duration: int, rms_threshold: int = 1000) -> bytes:
    """
    Fade out the audio data.

    :param data: the audio data to fade out
    :type data: bytes
    :param fade_duration: the duration of the fade in samples
    :type fade_duration: int
    :param rms_threshold: the threshold for determining if audio is present
    :type rms_threshold: int
    :return: the audio data with a fade-out applied
    :rtype: bytes
    """

    def rms(aud_data):
        return sqrt(maximum(mean(square(aud_data)), 0))

    num_samples = len(data) // 2  # Divide by 2 for 16-bit audio samples
    fade_samples = min(num_samples, fade_duration)
    audio_data = frombuffer(data, dtype=int16).copy()  # Create a writeable copy of the array

    start_fade = 0
    for i in range(0, num_samples - fade_samples, fade_samples):
        if rms(audio_data[i:i + fade_samples]) < rms_threshold:
            start_fade = i
            break

    fade = linspace(1, 0, num_samples - start_fade).astype(int16)  # Convert fade array to int16
    audio_data[start_fade:] *= fade
    return audio_data.tobytes()


def start_audio_stream(rate: int, length: int) -> None:
    """
    Start the audio stream.

    :param rate: the sampling rate of the audio stream
    :type rate: int
    :param length: the length of the audio stream in frames per buffer
    :type length: int
    """
    global audio_stream
    audio_stream = pa.open(rate=rate, channels=1, format=pyaudio.paInt16, input=True,
                           frames_per_buffer=length, input_device_index=None)


def stop_audio_stream() -> None:
    """
    Stop the audio stream.
    """
    global audio_stream
    if audio_stream is not None:
        try:
            audio_stream.stop_stream()
            audio_stream.close()
        except Exception as e:
            logger.error(f"Failed to stop the audio stream: {e}")
        finally:
            audio_stream = None


def get_next_audio_frame(handle):
    """
    Read the next frame from the audio stream.

    :param handle: The Porcupine handle
    :return: tuple, the unpacked PCM data
    """
    global audio_stream
    try:
        pcm = audio_stream.read(handle.frame_length, exception_on_overflow=False)
    except pyaudio.PyAudioError as e:
        logger.error(f"Input overflow error while reading from the audio stream: {e}")
        return None

    pcm = struct.unpack_from("h" * handle.frame_length, pcm)
    return pcm


def shutdown_audio() -> None:
    """
    Shutdown the audio stream and terminate the pyaudio instance.
    """
    global audio_stream
    global pa
    stop_audio_stream()
    pa.terminate()


atexit.register(shutdown_audio)



import time
import random
import requests
import scrapy
import os

headers = {
  "User-agent":
  "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.102 Safari/537.36 "
  "Edge/18.19582"
}


def simple_google_search(query, retries=2):
    params = {
      "q": query,
    }
    html = requests.get('https://www.google.com/search', headers=headers, params=params)
    # read the html content with scrapy
    content = scrapy.Selector(text=html.content)
    response = ""
    try:
        calc = get_calculator(content)
        response += f"Expression: {query}\nAnswer: {calc}\n"
    except:
        pass
    try:
        weather = get_weather(content)
        return weather
    except:
        pass
    try:
        population = get_population(content)
        if population.find(r'[0-9]') > -1:
            return f"Population: {population}\n"

    except:
        pass
    try:
        second_answer = get_second_answer(content)
        if len(second_answer) > 2 and second_answer[:3] != "Ad\n":
            return f'Answer: {second_answer}\n'
    except:
        pass
    try:
        answer = get_answer(content)
        if len(answer) > 2 and answer[:3] != "Ad\n":
            return f'Answer: {answer}\n'
    except:
        pass
    try:
        basic = get_answer_basic(content)
        if len(basic) > 2 and basic[:3] != "Ad\n":
            return f'Answer: {basic}\n'
    except:
        pass
    try:
        address = get_address(content)
        if len(address) > 2:
            return address
    except:
        pass
    try:
        conversion = get_conversion(content)
        if len(conversion) > 2:
            return conversion
    except:
        pass
    try:
        translation = get_translation(content)
        if len(translation) > 2:
           return f"Translation: {translation}\n"
    except:
        pass
    try:
        flights_one = get_flights_one(content)
        if flights_one.find(r'[0-9]') > -1:
            return f"Flights: {flights_one}\n"
    except:
        pass
    try:
        flights_two = get_flights_two(content)
        if flights_two.find(r'[0-9]') > -1:
            return f"Flights: {flights_two}\n"
    except:
        pass
    try:
        sports = get_sports(content)
        if sports.find(r'[0-9]') > -1:
            return f"Sports: {sports}\n"
    except:
        pass
    if retries > 0:
        time.sleep(4+random.random()*2)
        return simple_google_search(query, retries=retries-1)
    return "No simple answer found."


def test_google_search():
    tests = ["Address of the Fox Building", "2*2+2", "Baltimore weather", "London population", "100 usd in gbp",
             "swagger definition", "luke skywalker lightsaber color", "Who is using IBM", "hello in french",
             "Flights from Baltimore to Miami", "Flights from Baltimore to Miami on Tuesday", "nfl matches"]
    tests = random.sample(tests, len(tests))
    fails = []
    errors = []
    for query in tests:
        try:
            result = simple_google_search(query)
            time.sleep(4+random.random()*2)
            if result == "No simple answer found.":
                fails.append(query)
                print(f"Test failed: {query}")
            else:
                print(f"Test passed: {query}:{result}")
        except Exception as e:
            errors.append(query)
            print(f"Test error: {query} - {e}")
    if len(fails) == 0 and len(errors) == 0:
        print("All tests passed!")
        return
    else:
        raise Exception("simple_google_search tests failed.")


def open_search_tool(query):
    params = {
        "q": query,
    }
    html = requests.get('https://www.google.com/search', headers=headers, params=params)
    # read the html content with scrapy
    content = scrapy.Selector(text=html.content)
    with open("test.html", "w") as f:
        f.write(html.text)
    # open the file in a browser
    os.system("open test.html")
    return content


# Address of the Fox Building
def get_address(content):
    address = content.xpath('//*[@id="main"]/div[3]/div/div[3]/div/div/div/div/div/div/div/div/div/text()')[0].get()
    return address


# 2*2+2
def get_calculator(content):
    calc = content.xpath('//*[@id="main"]/div[3]/div/div[3]/div/div/div/div/div/div/div/div/div/text()')[0].get()
    return calc


# Baltimore weather
def get_weather(content):
    temperature = content.xpath('//*[@id="main"]/div[3]/div/div[3]/div/div/div/div/div[1]/div[1]/div/div/div/div/text()')[0].get()
    condition = content.xpath('//*[@id="main"]/div[3]/div/div[3]/div/div/div/div/div[1]/div[2]/div/div/div/div/text()')[0].get()
    condition = condition.replace("\u202f", "")
    return temperature + "\n" + condition


# London population
def get_population(content):
    population = content.xpath('//*[@id="main"]/div[3]/div/div[3]/div/div/div/div/div[1]/div[1]/div/div/div/div/text()')[0].get()
    return population


# 100 usd in gbp
def get_conversion(content):
    conversion = content.xpath('//*[@id="main"]/div[3]/div/div[3]/div/div/div/div/div[1]/div/div/div/div/text()')[0].get()
    return conversion


# swagger definition
def get_answer_basic(content):
    definition = "\n".join(content.xpath('//*[@id="main"]/div[3]/div/div[3]/div/div/*//text()').extract())
    return definition


# luke skywalker lightsaber color
def get_answer(content):
    answer = "\n".join(content.xpath('//*[@id="main"]/div[3]/div/div[2]/div/*//text()').extract())
    return answer


# Who is using IBM
def get_second_answer(content):
    answer = "\n".join(content.xpath('//*[@id="main"]/div[3]/div/div/*//text()').extract())
    return answer


# hello in french
def get_translation(content):
    translation_input_language = content.xpath('//*[@id="tsuid_2"]/option[contains(@selected, "selected")]/text()').get()
    translation_input = content.xpath('//*[@id="lrtl-source-text"]/input/@value').get()
    translation_output_language = content.xpath('//*[@id="tsuid_4"]/option[contains(@selected, "selected")]/text()').get()
    translation_output = content.xpath('//*[@id="lrtl-translation-text"]/text()').get()
    if translation_input_language is not None and translation_input is not None and \
        translation_output_language is not None and translation_output is not None:
        translation = f"{translation_input_language}: {translation_input}\n{translation_output_language}: {translation_output}"
        return translation
    else:
        return ""


# Flights from Baltimore to Miami
def get_flights_one(content):
    flights = "\n".join(content.xpath('//*[@id="main"]/div[4]/div/div/div[1]/*//text()').extract())
    return flights


# Flights from Baltimore to Miami on Tuesday
def get_flights_two(content):
    flights = "\n".join(content.xpath('//*[@id="main"]/div[3]/*//text()').extract())
    return flights


# nfl matches
def get_sports(content):
    sports = "\n".join(content.xpath('//*[@id="main"]/div[3]/div*//text()').extract())
    return sports